/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef NEST_CONCURRENCY_MANAGER_H
#define NEST_CONCURRENCY_MANAGER_H

#include "nest_common.h"
#include "nest_client.h"
#include "nest_concurrency_model.h"

class
nestConcurrencyManager {

public:

    static int init( 
        fd_set *readfds, fd_set *writefds, fd_set *r_ready, fd_set *w_ready,
        void( *notifyServerCallback )(nestClient *, NestReplyStatus ) );

    static void sigChild( int signal, siginfo_t *info, void *context );

    static void makeProgress( );

    static void taskDone(nestClient *client, NestReplyStatus status, int model);
    
    static NestReplyStatus addTask( nestClient *client );
    
    static int getMaxfd( ); 

    static char *getConcurrencyModel( );

    static void debugLevel( long level );
    
    static int processFunction( int id );

    static ssize_t writeAd( nestProtocol *ad );

private:

    static void setUpModelList( );

    static nestConcurrencyModel *selectConcurrencyModel( nestClient *client );
    static nestConcurrencyModel *selectUsingAverages( long transferlength );
    static nestConcurrencyModel *selectTakingTurns( );

    static nestClient *selectFavoriteProtocol( );
    static nestClient *selectShortestFile( );
    static nestClient *selectCachedFile();

    static nestClient *selectLowestStride( );
    static unsigned long findLowestStride( );
    static int makeSchedule();
    static NestReplyStatus makeStrides();
    static void strideDone( nestClient *, long len );

    static NestReplyStatus lockMutex();
    static NestReplyStatus unlockMutex();

    
    static nestClient *selectQueuedTask( );

    static NestReplyStatus startTask( nestClient *client );
    
    static void filecache_model_check( char *file, 
            nestFiletransfer *transfer, int model_id);

    static char * modelName( int model );
    
};


#endif
