/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_DEBUG_H
#define _NEST_DEBUG_H

#include "nest_common.h"
#include "nest_error.h"
#include "nest_protocol.h"

//** concurrencies
#define NEST_DEBUG_POP              0x00000001
#define NEST_DEBUG_POT              0x00000002
#define NEST_DEBUG_NOB              0x00000004
#define NEST_DEBUG_CONCURRENCY      0x0000000f

//** miscellaneous
#define NEST_DEBUG_FLOCK            0x00000010
#define NEST_DEBUG_MALLOC           0x00000020
#define NEST_DEBUG_FD_CACHE         0x00000040

//** account management
#define NEST_DEBUG_USER_MNGMNT      0x00000100
#define NEST_DEBUG_GROUP_MNGMNT     0x00000200
#define NEST_DEBUG_ACCT_MNGMNT      0x00000300

//** reservations
#define NEST_DEBUG_LOTS             0x00000400
#define NEST_DEBUG_BW_RSV           0x00000800
#define NEST_DEBUG_RSV              0x00000c00

//** more miscellaneous
#define NEST_DEBUG_ACL              0x00001000
#define NEST_DEBUG_FDSETS           0x00002000
#define NEST_DEBUG_DEBUG            0x00004000
#define NEST_DEBUG_TRANSFER         0x00008000

#define NEST_DEBUG_CLIENT           0x00010000
#define NEST_DEBUG_STORAGE          0x00020000
#define NEST_DEBUG_SERVER           0x00040000
#define NEST_DEBUG_ADMIN            0x00080000

#define NEST_DEBUG_AUTH             0x00100000
#define NEST_DEBUG_CACHEMODEL       0x00200000
#define NEST_DEBUG_FD_LEAKS         0x00400000
#define NEST_DEBUG_PROXY            0x00800000

//** protocols
#define NEST_DEBUG_GRIDFTP          0x01000000
#define NEST_DEBUG_CHIRP            0x02000000
#define NEST_DEBUG_HTTP             0x04000000
#define NEST_DEBUG_FS               0x08000000
#define NEST_DEBUG_NFS              0x10000000
#define NEST_DEBUG_PROTOCOL_SCHED   0x20000000
#define NEST_DEBUG_PROTOCOL         0xff000000

#define NEST_DEBUG_ALWAYS           0xffffffff
#define NEST_DEBUG_NEVER            0x00000000

// used to turn debugging on or off
void set_nest_debug_level( long level );
void add_nest_debug_level( long new_level );
NestReplyStatus add_nest_debug_level( const char *debug );
bool check_nest_debug_status( long level );

NestReplyStatus nest_debug_init( nestProtocol *, bool daemonize );

long get_nest_debug_level( );

int nest_debug( long level, const char *fmt, ...);
void nest_perror( const char *msg);
int nest_printf( const char *fmt, ... ); 
int nest_warn( const char *filename, const char *func_name, const char *fmt, ... );
void nest_die( const char *fmt, ... );

#endif
