/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef NEST_FILECACHE_MODEL_INFINITE_H
#define NEST_FILECACHE_MODEL_INFINITE_H

#include <stdio.h>
#include <assert.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "nest_common.h"
// #include "nest_filecache_model_simple.h"
#include "nest_hash_table.h"

class nestFilecacheModelInfinite : public nestFilecacheModel {

public:

    nestFilecacheModelInfinite( int hashsize );

    virtual NestReplyStatus insert( bool *hit, const char *file, long size );

    virtual bool remove( const char *file );
    
    virtual bool check( const char *file );

    virtual void debug( );

    virtual bool grow( float megs );

    virtual bool shrink( float megs );

    virtual bool renameFile( const char *oldname, const char *newname );

    virtual float hit_rate( );
    
    virtual long gethits( );

    virtual long getmisses( );
    
    virtual long getevictions( );
    
    virtual float getsize( );

    virtual long getavail( );

    virtual long getentries( );
    
    virtual long getFilesize( const char *file );

    virtual void setFilesize( const char *file, long size );

    virtual void writeAd( nestProtocol *ad );

    virtual NestReplyStatus format();

    virtual bool rememberLength( const char *file, long *length );
    
private:
    struct nest_hash_table *infinite_cache;
    unsigned long hits, misses; 
    long size;
};

#endif
