/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_FILETRANSFER_H 
#define _NEST_FILETRANSFER_H 

#include "nest_common.h"
#include "nest_protocol.h"
#include "network_limit.h"
#include "nest_request.h"

struct IO_Measurement {
    unsigned long total_bytes;

    unsigned long total_in_time;
    unsigned long total_out_time;

    unsigned int  first_in_time;
    unsigned int  first_out_time;
    unsigned int  first_in_size;
    unsigned int  first_out_size;
};

class
nestFiletransfer {

public:

    nestFiletransfer( nestProtocol *src, nestProtocol *dst, long length, 
            char *buffer, long bufsize, int type ); 

    ~nestFiletransfer( );

    bool isIncoming();
    
    void setLimit( long allocation );
        
    long getLimit( );
    
    NestReplyStatus BlockingTransfer( );
    
    NestReplyStatus startTransfer( );

    NestReplyStatus endTransfer( NestReplyStatus status );

    nestProtocol *getSrc( ) { return src; }

    nestProtocol *getDst( ) { return dst; }
    
    struct IO_Measurement *getMeasurements();
    
    void setMeasurements( struct IO_Measurement *measurements );

    long getLength( ) { return initlen; }

    long getInitialLength( );
    
    long getBytesTransferred( );

    void setBytesTransferred( long bytes );

    float getBandwidth( );

    long getLatency( );

    int getType( );

    void setCacheGuess( bool b ) { cache_guess = b; }

    bool getCacheGuess( ) { return cache_guess; }

    void setLengthGuess( long len ) { len_guess = len; }

    long getLengthGuess( ) { return len_guess; }

private:

    nestProtocol *src;
    nestProtocol *dst;
    char *buffer;
    long buflength;
    long initlen;
    long allocation;
    struct timeval start, *end;
    int type;
    struct IO_Measurement info; 
    long len_guess;
    bool cache_guess;
    bool started;
};

#endif



