/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_globus.h"


void
printGSS_Error( OM_uint32 major_status, OM_uint32 minor_status ) {
    char *error_string = NULL;
    globus_gss_assist_display_status_str( 
            &error_string,
            NULL,
            major_status,
            minor_status,
            0 );
    nest_debug( NEST_DEBUG_AUTH, "Unable to authenticate: %s.\n", 
            error_string );
    if ( error_string != NULL ) {
        NestFree( error_string );
    }
}

NestReplyStatus
inquireGlobusCredential(    gss_name_t *name, 
                            OM_uint32 *lifetime,
                            gss_cred_id_t *credential_handle ) 
{
    OM_uint32 major_status; 
    OM_uint32 minor_status; 
    gss_OID_set mechanisms;
    gss_cred_usage_t cred_usage;

    major_status = gss_inquire_cred(
                        &minor_status,
                        credential_handle,
                        name,
                        lifetime,
                        &cred_usage,
                        &mechanisms );

    if ( major_status != GSS_S_COMPLETE ) {
        printGSS_Error( major_status, minor_status );
        return NEST_UNKNOWN_USER;
    } else {
        return NEST_SUCCESS;
    }

}

NestReplyStatus 
acquireGlobusCredential( gss_cred_id_t *credential_handle, 
gss_cred_usage_t credential_usage ) 
{
    OM_uint32 major_status; 
    OM_uint32 minor_status; 

    /*
    major_status = globus_gss_assist_acquire_cred_ext( &minor_status,
                    0,
                    GSS_C_INDEFINITE, 
                    GSS_C_NO_OID_SET,
                    GSS_C_INITIATE, 
                    credential_handle,
                    NULL, NULL);
                */

    major_status = globus_gss_assist_acquire_cred(&minor_status,
                credential_usage, 
                credential_handle); 

    if ( major_status != GSS_S_COMPLETE ) {
        printGSS_Error( major_status, minor_status );
        return NEST_UNKNOWN_USER; 
    } else {
        return NEST_SUCCESS;
    }
}


/** taken and modified from Doug Thain's Kangaroo server */
int 
auth_globus_barrier( FILE *r, FILE *w, int value ) {
    int fields, response;
    char line[AUTH_LINE_MAX];

    fprintf(w,"%d\n",value);
    fflush(w);

    if(fgets(line,AUTH_LINE_MAX,r)) {
        fields = sscanf(line,"%d",&response);
        if(fields==1) {
            return response;
        }
    }

    return 0;
}

int
makeFileStars( FILE **r, FILE **w, int socket ) {
    int retval;
    *r = fdopen( socket, "rb" );
    *w = fdopen( socket, "wb" );
    if ( !(*r) || !(*w) ) {
        nest_die( "Couldn't convert socket %d into FILE *'s\n" );
        retval = 0;
    } else {
        retval = 1;
    }
    return retval;
}

/** taken and modified from Doug Thain's Kangaroo server */
// this is what the client calls
int 
auth_globus_assert( int sd, char *server ) {

    gss_cred_id_t credential = GSS_C_NO_CREDENTIAL;
    gss_ctx_id_t context = GSS_C_NO_CONTEXT;
    OM_uint32 major, minor, flags=0;
    int token;
    int result;
    char *my_server = NULL;
    int socket = NestDup( sd );

    FILE *r, *w;

    if ( ! makeFileStars( &r, &w, socket ) ) {
        result = 0;
        goto done;
    }

    //major = globus_gss_assist_acquire_cred(&minor,GSS_C_INITIATE,&credential);
    major = globus_gss_assist_acquire_cred_ext( &minor,
                    0,
                    GSS_C_INDEFINITE, 
                    GSS_C_NO_OID_SET,
                    GSS_C_INITIATE, 
                    &credential,
                    NULL, NULL);
    if( major!=GSS_S_COMPLETE ) {
        nest_warn( __FILE__, __FUNCTION__, 
                "auth_globus: couldn't load my credentials: "
                "did you run grid-proxy-init?\n");
        result = 0;
    } else {
        result = 1;
    }

    if( ! auth_globus_barrier( r,w,result ) ) {
        nest_debug( NEST_DEBUG_AUTH, 
                "auth_globus: other side couldn't load credentials\n");
        result = 0;
        goto done;
    }

    if( ! result ) {
        goto done;
    }

    nest_debug( NEST_DEBUG_AUTH, "auth_globus: sending credentials\n");

    setbuf(w,0);
    if ( server != NULL ) {
        my_server = (char *)NestMalloc( 
            NestStrlen( server ) + NestStrlen( "host@" ) + 1 );
        sprintf( my_server, "host@%s", server );
    } else {
        my_server = NULL;
    }

    // TODO maybe add delegate to the flags parameter
    // see http://www.globus.org/security/gss_assist.html
    // or maybe in the ncftp code
    major = globus_gss_assist_init_sec_context( 
            &minor, 
            credential, 
            &context, 
            my_server, 
            0, 
            &flags, 
            &token,
            globus_gss_assist_token_get_fd, 
            r, 
            globus_gss_assist_token_send_fd, 
            w );
    if( major != GSS_S_COMPLETE ) {
        char *error_string = NULL;
        globus_gss_assist_display_status_str( 
                &error_string,
                NULL,
                major,
                minor,
                token );
        nest_debug( NEST_DEBUG_AUTH, "Unable to authenticate: %s.\n", 
                error_string );
        result = 0;
        goto done;
    }

    nest_debug( NEST_DEBUG_AUTH, "auth_globus: credentials accepted\n");
    result = 1;

    done:

    if ( my_server != NULL ) NestFree( my_server );
    setvbuf(w,0,_IOFBF,BUFSIZ);
    gss_delete_sec_context( &minor, &context, GSS_C_NO_BUFFER );
    gss_release_cred( &major, &credential );
    NestClose( socket );
    fclose( r );
    fclose( w );

    return result;
}

/** taken and modified from Doug Thain's Kangaroo server */
// this is what the server calls
// TODO: March 11, 2003.  This hung and froze a server.  Why?
NestReplyStatus 
auth_globus_accept( int sd, char **subject, char **globus_username ) {

    static gss_cred_id_t credential = GSS_C_NO_CREDENTIAL;
    OM_uint32 minor = 0; 
    OM_uint32 major = 0; 
    gss_ctx_id_t context = GSS_C_NO_CONTEXT;
    OM_uint32 flags=0;
    int token;
    NestReplyStatus result = NEST_SUCCESS;
    char *globusname = NULL;
    FILE *r, *w;
    int ret = 1;
    int socket = NestDup( sd ); // why do the dup?  I don't know
                                // but if don't do the dup, then the fd
                                // gets bad.  I think probably the fd
                                // is implicitly closed when we do the
                                // fcloses on the FILE * that we make from it
    int user_user = 0;

    //LOCK_MUTEX( &nest_common_mutex );

        // only get initial credential once 
    if ( credential == GSS_C_NO_CREDENTIAL ) {
        major = globus_gss_assist_acquire_cred( &minor, GSS_C_INITIATE,
                &credential);
        if( major!=GSS_S_COMPLETE ) {
			char	*str = NULL;
			globus_gss_assist_display_status_str( &str, NULL, major, minor, 0 );
            nest_warn( __FILE__, __FUNCTION__,
                    "Could not get initial globus credential.  "
                    "Did you run grid-proxy-init?\n%s\n", str );
            result = NEST_NONPRIVELEDGED_SERVER;
        } else {
            result = NEST_SUCCESS;
        }
    }

    if ( ! makeFileStars( &r, &w, socket ) ) {
        result = NEST_UNKNOWN_ERROR;
    }

        // need to do barrier even if could not acquire credential
    if ( ! auth_globus_barrier( r, w, ret ) ) {
        nest_debug( NEST_DEBUG_AUTH, 
                "auth_globus: other side couldn't load credentials\n");
        result = NEST_COULD_NOT_AUTHENTICATE;
        goto done;
    }

        // if anything did not work then don't try to receive credential
        // because client will not be sending one
    if( result != NEST_SUCCESS || ! ret ) {
        goto done;
    }

    nest_debug( NEST_DEBUG_AUTH, "auth_globus: receiving credentials\n");

    setbuf(w,0);

        // this here is the globus line that froze
    major = globus_gss_assist_accept_sec_context( 
            &minor, 
            &context, 
            credential, 
            &globusname, 
            &flags, 
            &user_user, 
            &token, 
            0, 
            globus_gss_assist_token_get_fd, 
            r, 
            globus_gss_assist_token_send_fd, 
            w );
    if( major != GSS_S_COMPLETE ) {
        char *error_string = NULL;
        globus_gss_assist_display_status_str( 
                &error_string,
                NULL,
                major,
                minor,
                token );
        nest_debug( NEST_DEBUG_AUTH, "Unable to authenticate: %s.\n", 
                error_string );
        result = NEST_COULD_NOT_AUTHENTICATE;
        goto done;
    }

    nest_debug( NEST_DEBUG_AUTH, 
            "auth_globus (user_user %d): remote user is '%s'\n",
                user_user, globusname);
    result = NEST_SUCCESS;
    *globus_username = NestStrdup( globusname );

        // try to look it up in a grid-map file
    if ( globus_gss_assist_gridmap( globusname, subject ) != 0 ) {
            // failed to look it up
        if ( user_user ) {
                // its ok if we can't look it up
            *subject = NestStrdup( globusname );
            nest_debug( NEST_DEBUG_AUTH, 
                    "auth_globus: user2user auth, user is %s\n", 
                    *subject);
            result = NEST_SUCCESS;
        } else {
            result = NEST_USER_NOT_FOUND;
        }
    } else {
            // success
        result = NEST_SUCCESS;
        nest_debug( NEST_DEBUG_AUTH, "auth_globus: user is %s\n", *subject);
    }

    done:
    setvbuf(w,0,_IOFBF,BUFSIZ);
    gss_delete_sec_context( &minor, &context, GSS_C_NO_BUFFER );
    if ( globusname != NULL ) NestFree( globusname );
    NestClose( socket );
    fclose( r );
    fclose( w );
    
        // don't release initial credential - we can reuse it
    gss_release_cred( &major, &credential );
    //UNLOCK_MUTEX( &nest_common_mutex );

    return result;
}
