/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_PROTOCOL_HTTP_H
#define _NEST_PROTOCOL_HTTP_H

#include "nest_common.h"
#include "nest_protocol.h"
#include "nest_storage_manager.h"

#define HTTPMESSAGESIZE 4096

class nestProtocolHttp : public nestProtocol {
public:
    static int listen( int port );

    static nestProtocol* accept( int socket, sockaddr_in *address );
    
    static NestReplyStatus initiateThirdParty( nestClient *client,
            nestProtocol **connection );
    
    nestProtocolHttp( int socket );
    
    virtual ~nestProtocolHttp();

    virtual bool sendFinalAck( int transferType );
    
    virtual bool persistent();
    
    virtual bool sendAck(  );

    virtual bool sendAd( nestAd *ad );
    
    virtual bool sendTransferInfo( nestRequest* request );
    
    virtual bool sendErrorMessage(  const char* filename, const char* func_name, int error );
    
    virtual bool sendFilesize( long size);

    virtual nestRequest* receiveRequest(  fd_set *ready );

    virtual ssize_t Receive( char *buffer, ssize_t n );

    virtual ssize_t Send( const char *buffer, ssize_t n );
    
    virtual int getType( ) { return HTTP_PROTOCOL; }

    virtual bool hasSocket();

    virtual nestProtocol *clone();

    int getSocket();
    
    virtual bool endThirdPartyTransfer( int ); 

private:
    int socket;
    bool keep_alive;
    bool still_alive;
    bool already_closed;
    char message[HTTPMESSAGESIZE];

    static bool fullMessage( char *message ); 
    static NestReplyStatus HttpStatusToNestStatus( int status ); 

    nestRequest* closeConnectionRequest();
    char *convertFilename( char *req, char *filename, int BUFSIZE );
    
    bool closeNonpersistent();

    char *errorstring( int errorcode );
    
    bool sendMessage(  const void *vptr, int errorcode );

    bool sendHeader( int len, int errorcode );
};

#endif
