
/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#include "nest_proxy_client_cache.h"

    
nestProxyClientCache::nestProxyClientCache( ) {
    connections = nest_hash_table_create( 1037, NULL );
}

int destroyPool( const char *key, void *item, void *args ) {
    assert( args == NULL );
    nest_debug(NEST_DEBUG_PROXY, "Tearing down client cache pool for %s\n",key);
    delete( (nestClientPool*)item );
    return 1;
}
    
nestProxyClientCache::~nestProxyClientCache( ) {
    nest_hash_table_iterate( connections, destroyPool, NULL );
    nest_hash_table_delete( connections );
}

int sumPools( const char *key, void *item, void *args ) {
    int this_pool = ((nestClientPool*)item)->size();
    int *sum = (int *)args;
    *sum += this_pool;
    return 1;
}

int
nestProxyClientCache::size( ) {
    int sum = 0;
    nest_hash_table_iterate( connections, sumPools, &sum );
    return sum;
}


NestReplyStatus
nestProxyClientCache::insertProxyClient( const char *location, 
        nestClient *client ) 
{
    nestClientPool *clients = 
        (nestClientPool *)nest_hash_table_lookup( connections, location );
    if ( ! clients ) {
        nest_debug( NEST_DEBUG_PROXY, "Creating new client pool to cache"
                " proxy connections to %s.\n", location );
        clients = new nestClientPool( false );
        assert( clients );
        nest_hash_table_insert( connections, location, (void*)clients );
    }
	nest_debug( NEST_DEBUG_PROXY, "Inserting cached proxy to %s (%d)\n",
			location, client->getId() );
    clients->addClient( client );
    return NEST_SUCCESS;
}
    
nestClient *
nestProxyClientCache::getCachedProxy( const char *location ) 
{
    nestClientPool *clients = 
        (nestClientPool *)nest_hash_table_lookup( connections, location );

    if ( ! clients ) {
        return NULL;
    } else {
		nestClient *client = removeAvailableConnection( clients );
        if ( client != NULL ) {
            nest_debug( NEST_DEBUG_PROXY, "Returning cached proxy to %s (%d)\n",
				location, client->getId() );
        }
		return client;
    }
}

nestClient *
nestProxyClientCache::removeAvailableConnection( nestClientPool *clients ) {

    nestClient *client = NULL;
    client = clients->getFront();
    if ( client != NULL ) assert( ! client->isActive() );
/*
    for( CLIENT_ITERATOR itr=clients->begin(); itr!=clients->end(); *itr++) {
        nestClient *cursor = *itr;
        if ( ! cursor->isActive() ) {
            client = cursor;
            break;
        }
    }
*/
    return client;
}
