
/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_PROXY_CLIENT_CACHE_H
#define _NEST_PROXY_CLIENT_CACHE_H

#include "nest_common.h"
#include "nest_client_pool.h"
#include "nest_hash_table.h"
#include "nest_proxy.h"

// this class just maps strings (machine names) to colletions of 
// client objects to that machine.  it assumes that is it used
// synchronously and uses the internal isBusy() flag of a client
// to know whether a client is already being used for a transfer
// it only returns clients that are ! isBusy()
//
// this means this class is definitely not thread safe and is meant to be
// used in the syncronous serialized request handling thread of the dispatcher
// it also looks for not busy clients just by doing a list traversal which
// should be fine for small numbers of clients as in our expected workloads

class
nestProxyClientCache {

public:
    nestProxyClientCache( );

    ~nestProxyClientCache( );

        // if it returns NULL, you need to make one and insert it
        // it removes it from its internal list, when done you need to reinsert
    nestClient * getCachedProxy( const char *location );

    NestReplyStatus insertProxyClient( const char *location, nestClient *client );

    int size();

private:

    struct nest_hash_table *connections;

    nestClient *removeAvailableConnection( nestClientPool *clients );

};

#endif
