/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent, 
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 * 
 * For contact information, please visit: 
 *      http://www.cs.wisc.edu/condor/nest/
 *
 ***********************************************************************/

#ifndef _NEST_REQUEST_H
#define _NEST_REQUEST_H

#include "nest_common.h"

    // requests
enum {
    TIME_REQUEST,                       // 0
    CLIENT_SEND_FILE,
    CLIENT_RECV_FILE,
    FILE_DELETE_REQUEST,
    BANDWIDTH_STATISTICS_REQUEST,
    DIR_LIST_REQUEST,                   // 5
    MKDIR_REQUEST,
    CWD_REQUEST,
    PWD_REQUEST,
    FILESIZE_REQUEST,
    SERVER_TERMINATE_REQUEST,           // 10
    CLIENT_TERMINATE_REQUEST,
    TEST_MESSAGE,
    TEST_CLIENT_TO_SERVER,
    TEST_SERVER_TO_CLIENT,
    REQUEST_DONE,                       // 15
    CHANGE_USER_REQUEST,
    AVAILABLE_SPACE_QUERY,
    FORMAT_DISK_REQUEST,
    CONNECTION_REQUEST,
    CLOSE_CONNECTION_REQUEST,           // 20
    THIRD_PARTY_GET,
    THIRD_PARTY_PUT,
    DEBUG_REQUEST,
    RMDIR_REQUEST,
    NAME_COLLECTION,                    // 25
    UNNAME_COLLECTION,
    QUERY_COLLECTION,
    HELP_REQUEST,
    RHELP_REQUEST,
    NEST_SYNC,                          // 30
    REQUEST_NEST_AD,
    LINK_REQUEST,
    ADD_NEST_USER,
    ADD_NEST_GROUP,
    DELETE_NEST_USER,                   // 35
    BAD_REQUEST,
    DELETE_NEST_GROUP,
    ADD_USER_TO_GROUP,
    DELETE_USER_FROM_GROUP,
    REQUEST_USER_LOT,                   // 40
    REQUEST_GROUP_LOT,
    LIST_USER_LOTS,
    LIST_GROUP_LOTS,
    GET_LOT_STAT,
    UPDATE_LOT,                         // 45
    CHANGE_LOT_T,
    TERMINATE_LOT,
    VERSION_REQUEST,
    PHYSNAME_REQUEST,
    GET_ACL,                            // 50
    SET_ACL,                            
    GET_GROUP_ACL,
    SET_GROUP_ACL,
    LOT_MOVE_FILE,
    GET_FILEHANDLE,                     // 55
    FILEINFO_REQUEST,                   
    PROXY_GET,
    PROXY_PUT,
    PROXY_REGISTER,
    PROXY_QUERY,                        // 60
    COPY_REQUEST,                       
    ADD_USER_TO_LOT,
    DELETE_USER_FROM_LOT,
    FILE_RENAME_REQUEST,
    ATTACH_LOT_TO_FILE,                 // 65
    DETACH_LOT_FROM_FILE,
    PROXY_INDEX_PUSH,                   
    PROXY_FLUSH,
    SETATTR_REQUEST,
	SET_COLLECTION_MTIME,				// 70
	SET_COLLECTION_VERSION,
	SET_COLLECTION_COMMENT,
	GET_COLLECTION_MTIME,
	GET_COLLECTION_VERSION,
	GET_COLLECTION_COMMENT,
};

class
nestRequest {

public:

    nestRequest( int type,
                 char *nestfile,
                 char *arg1,
                 char *arg2,
                 long start,
                 long end,
                 char *arg3 = NULL,
                 char *arg4 = NULL,
                 char *arg5 = NULL,
                 char *arg6 = NULL );

	nestRequest( const nestRequest *req );
	
    ~nestRequest( );

    int getType( );

	bool isWrite();
	
	static bool isWrite( int type );
	
    char *toString();
    
    bool isThirdParty( );

    long getStart( );

//    long getOldStart( );
    
    long getEnd( );

    long getLength( ) { return end - start; }

    int  getAuthType( );

    NestReplyStatus  getErrorCode( );

    void setEnd( long end );

    void setStart( long start );

    void setStatus( NestReplyStatus status ) { this->status = status; }

    NestReplyStatus getStatus( ) { return this->status; }
    
    int getValue( );

    char *getProtocol( );
    
    const char *getNestFile( );

    const char *getOldNestFile( );
    
    const char *getNewNestFile( );

    const char *getProxyLocation( );

    NestReplyStatus setNestFile( const char *file ); 

    //void renameNestFile( char *file );

    //void restoreNestFileName( );

 //   const char *getOldNestFile( );
    
    const char *getUser( );

    const char *getGroup( );

    const char *getIdentity( );

    const char *getTargetUser( );

    const char *getTargetAcl( );

    bool getAclIsGroup( );

    int getLotType( );

    int getNewLotType( );

    double getLotSize( );

    double getLotDuration( );

    char *getLotUser( );
    
    char *getParentLotUser( );

    char *getParentLotGroup( );

    rsv_id_t getParentLot( );

    int  getLotHandle(NestLotHandle *handle);

    int getSourceLotHandle(NestLotHandle *handle);

    int getDestLotHandle(NestLotHandle *handle);

    double getLotMoveBytes( );

    const char *getCollectionName( );
	const char *getCollectionVersion( );
	const char *getCollectionComment( );
	time_t getCollectionMtime( );

    char *getLinkName( );

    char *getRemoteFile( );

    char *getRemoteHost( );

    char *getDebugValue( );

    void setFilehandle( void *fh );

    void *getFilehandle();

    void clearFilehandle( );
    
    void setFileattr( void *fattr );
    
    void clearFileattr( );
    
    void *getFileattr( );

    long getChunk();    // proxys need this

    void setChunk( long chunk );

    long getAge();      // returns in millis
    
private:
	void init(   int type,
                 char *nestfile,
                 char *arg1,
                 char *arg2,
                 long start,
                 long end,
                 char *arg3 = NULL,
                 char *arg4 = NULL,
                 char *arg5 = NULL,
                 char *arg6 = NULL );
    int type;
    long start;
    long end;
    long chunk;
//    long oldstart;
    char *nestfilename;
//    char *oldnestfilename;
    char *arg1, *arg2, *arg3, *arg4, *arg5, *arg6;
    char *handle;
    NestReplyStatus status;
    struct timeval createtime, starttime, endtime;
    void *fh;   // a way for nfs to pass filehandles around
    void *fattr;    // a way for nfs to pass fileattributes around
};

#endif
