/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 * This particular file was written by Muthian Sivathanu
 ***********************************************************************/



#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <rpc/pmap_clnt.h>
#include <rpc/svc.h>
#include <string.h>
#include <memory.h>
#include <sys/socket.h>
#include <linux/types.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <rpc/svc.h>
#include "rpcudp_lib.h"

#include "nest_debug.h"

#ifndef SIG_PF
#define SIG_PF void(*)(int)
#endif


struct xp_ops my_svcudp_op = {
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
};

#define rpc_buffer(xprt) ((xprt)->xp_p1)

/*
 * kept in xprt->xp_p2
 */
struct my_svcudp_data
  {
    u_int su_iosz;        /* byte size of send.recv buffer */
    u_long su_xid;        /* transaction id */
    XDR su_xdrs;        /* XDR handle */
    char su_verfbody[MAX_AUTH_BYTES];    /* verifier body */
    char *su_cache;        /* cached data, NULL if no cache */
  };

#define    su_data(xprt)    ((struct my_svcudp_data *)(xprt->xp_p2))



SVCXPRT *
my_svcudp_bufcreate (int sock, u_int sendsz, u_int recvsz)
{
  bool_t madesock = FALSE;
  SVCXPRT *xprt;
  struct my_svcudp_data *su;
  struct sockaddr_in addr;
  socklen_t len = sizeof (struct sockaddr_in);
  int pad;

  if (my_svcudp_op.xp_recv == NULL) {
      printf("\nsvcudp_ops not yet initialized! cannot create udp xprt !");
    return NULL;
  }

  if (sock == RPC_ANYSOCK)
    {
      if ((sock = socket (AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
      printf ("my_svcudp_create: socket creation problem"); fflush(stdout);
      return (SVCXPRT *) NULL;
    }
      madesock = TRUE;
    }
  __bzero ((char *) &addr, sizeof (addr));
  addr.sin_family = AF_INET;
  if (bindresvport (sock, &addr))
    {
      addr.sin_port = 0;
      (void) bind (sock, (struct sockaddr *) &addr, len);
    }
  if (getsockname (sock, (struct sockaddr *) &addr, &len) != 0)
    {
      printf ("my_svcudp_create - cannot get sockname for %d: %s\n",
              sock, strerror( errno )); 
      fflush(stdout);
      if (madesock)
    (void) close (sock);
      return (SVCXPRT *) NULL;
    }
  xprt = (SVCXPRT *) mem_alloc (sizeof (SVCXPRT));
  if (xprt == NULL)
    {
      (void) fputs ("my_svcudp_create: out of memory\n", stderr);
      return NULL;
    }
  su = (struct my_svcudp_data *) mem_alloc (sizeof (*su));
  if (su == NULL)
    {
      (void) fputs ("my_svcudp_create: out of memory\n", stderr);
      return NULL;
    }
  su->su_iosz = ((MAX (sendsz, recvsz) + 3) / 4) * 4;
  if ((rpc_buffer (xprt) = (char*)mem_alloc (su->su_iosz)) == NULL)
    {
      (void) fputs ("my_svcudp_create: out of memory\n", stderr);
      return NULL;
    }
  xdrmem_create (&(su->su_xdrs), rpc_buffer (xprt), su->su_iosz, XDR_DECODE);
  su->su_cache = NULL;
  xprt->xp_p2 = (caddr_t) su;
  xprt->xp_verf.oa_base = su->su_verfbody;
  (xp_ops*)xprt->xp_ops = &my_svcudp_op;
  xprt->xp_port = ntohs (addr.sin_port);
  xprt->xp_sock = sock;

#ifdef IP_PKTINFO
  if (setsockopt (sock, SOL_IP, IP_PKTINFO, (void *) &pad,
          sizeof (pad)) == 0)
    /* Set the padding to all 1s. */
    pad = 0xff;
#endif

  return xprt;
}


SVCXPRT *
my_svcudp_create (int sock)
{
  return my_svcudp_bufcreate (sock, UDPMSGSIZE, UDPMSGSIZE);
}


void
my_svcudp_destroy (SVCXPRT *xprt, int to_close)
{
  struct my_svcudp_data *su = su_data (xprt);

  //cout <<"\nsvcudp_destroy called - transp addr = " <<(int)xprt <<endl;

  if (to_close)
      (void) close (xprt->xp_sock);
  XDR_DESTROY (&(su->su_xdrs));

  mem_free (rpc_buffer (xprt), su->su_iosz);
  mem_free ((caddr_t) su, sizeof (struct my_svcudp_data));
  mem_free ((caddr_t) xprt, sizeof (SVCXPRT));
}


/*
 * Unix longhand authenticator
 *
 * copied from glib-c/sunrpc/svc_authux.c
 */
enum auth_stat
_svcauth_unix (struct svc_req *rqst, struct rpc_msg *msg)
{
  enum auth_stat stat;
  XDR xdrs;
  struct authunix_parms *aup;
  int32_t *buf;
  struct area
    {
      struct authunix_parms area_aup;
      char area_machname[MAX_MACHINE_NAME + 1];
      gid_t area_gids[NGRPS];
    }
   *area;
  u_int auth_len;
  u_int str_len, gid_len;
  u_int i;

  area = (struct area *) rqst->rq_clntcred;
  aup = &area->area_aup;
  aup->aup_machname = area->area_machname;
  aup->aup_gids = area->area_gids;
  auth_len = (u_int) msg->rm_call.cb_cred.oa_length;
  xdrmem_create (&xdrs, msg->rm_call.cb_cred.oa_base, auth_len, XDR_DECODE);
  buf = XDR_INLINE (&xdrs, auth_len);
  if (buf != NULL)
    {
      aup->aup_time = IXDR_GET_LONG (buf);
      str_len = IXDR_GET_U_INT32 (buf);
      if (str_len > MAX_MACHINE_NAME)
	{
	  stat = AUTH_BADCRED;
	  goto done;
	}
      bcopy ((caddr_t) buf, aup->aup_machname, (u_int) str_len);
      aup->aup_machname[str_len] = 0;
      str_len = RNDUP (str_len);
      buf = (int32_t *) ((char *) buf + str_len);
      aup->aup_uid = IXDR_GET_LONG (buf);
      aup->aup_gid = IXDR_GET_LONG (buf);
      gid_len = IXDR_GET_U_INT32 (buf);
      if (gid_len > NGRPS)
	{
	  stat = AUTH_BADCRED;
	  goto done;
	}
      aup->aup_len = gid_len;
      for (i = 0; i < gid_len; i++)
	{
	  aup->aup_gids[i] = IXDR_GET_LONG (buf);
	}
      /*
       * five is the smallest unix credentials structure -
       * timestamp, hostname len (0), uid, gid, and gids len (0).
       */
      if ((5 + gid_len) * BYTES_PER_XDR_UNIT + str_len > auth_len)
	{
	  (void) printf ("bad auth_len gid %d str %d auth %d\n",
			 gid_len, str_len, auth_len);
	  stat = AUTH_BADCRED;
	  goto done;
	}
    }
  else if (!xdr_authunix_parms (&xdrs, aup))
    {
      xdrs.x_op = XDR_FREE;
      (void) xdr_authunix_parms (&xdrs, aup);
      stat = AUTH_BADCRED;
      goto done;
    }

  /* get the verifier */
  if ((u_int)msg->rm_call.cb_verf.oa_length)
    {
      rqst->rq_xprt->xp_verf.oa_flavor =
	msg->rm_call.cb_verf.oa_flavor;
      rqst->rq_xprt->xp_verf.oa_base =
	msg->rm_call.cb_verf.oa_base;
      rqst->rq_xprt->xp_verf.oa_length =
	msg->rm_call.cb_verf.oa_length;
    }
  else
    {
      rqst->rq_xprt->xp_verf.oa_flavor = AUTH_NULL;
      rqst->rq_xprt->xp_verf.oa_length = 0;
    }
  stat = AUTH_OK;
done:
  XDR_DESTROY (&xdrs);
  return stat;
}

void
my_svc_getreq_common (svc_dispatch disp_fn, const int fd )
{
  enum xprt_stat stat;
  struct rpc_msg msg;
  register SVCXPRT *xprt;
  char cred_area[2 * MAX_AUTH_BYTES + RQCRED_SIZE];

  if ( ! FD_ISSET( fd, &svc_fdset ) ) {
      nest_debug( NEST_DEBUG_NFS, 
              "svc_getreq_common ignoring notmy fd %d\n", fd );
      return;
  }

  msg.rm_call.cb_cred.oa_base = cred_area;
  msg.rm_call.cb_verf.oa_base = &(cred_area[MAX_AUTH_BYTES]);

  xprt = my_svcudp_create(fd);
  /* Do we control fd? */
  if (xprt == NULL)
     return;

  //fprintf( stderr, "svc_getreq_common reading from %d\n", fd );

  /* now receive msgs from xprtprt (support batch calls) */
  do
    {
      if (SVC_RECV (xprt, &msg))
    {
      /* now find the exported program and call it */
      //struct svc_callout *s;
      struct svc_req r;
      enum auth_stat why;
      //int prog_found;

      r.rq_clntcred = &(cred_area[2 * MAX_AUTH_BYTES]);
      r.rq_xprt = xprt;
      r.rq_prog = msg.rm_call.cb_prog;
      r.rq_vers = msg.rm_call.cb_vers;
      r.rq_proc = msg.rm_call.cb_proc;
      r.rq_cred = msg.rm_call.cb_cred;

      /* first authenticate the message */
      /* Check for null flavor and bypass these calls if possible */

      // muthian took out a bunch of code here . . . to make it threadsafe

        // from glibc/sunrpc/svc.c, put some of it back in
        // leave out the stuff about finding the correct service ....
        if (msg.rm_call.cb_cred.oa_flavor == AUTH_NULL)
        {
            r.rq_xprt->xp_verf.oa_flavor = _null_auth.oa_flavor;
            r.rq_xprt->xp_verf.oa_length = 0;
        }
        else if ((why = _svcauth_unix (&r, &msg)) != AUTH_OK)
        {
            svcerr_auth (xprt, why);
            goto call_done;                                                       
        }

//      r.rq_xprt->xp_verf.oa_flavor = _null_auth.oa_flavor;
//      r.rq_xprt->xp_verf.oa_length = 0;

      (*disp_fn)(&r, xprt);
      goto call_done;
    }
    call_done:
        ;
/*      if ((stat = SVC_STAT (xprt)) == XPRT_DIED)
    {
      cout <<"\nrpclib_udp: DESTROYING xprt !"<<endl;
      SVC_DESTROY (xprt);
      break;
    } */
    }
  while (stat == XPRT_MOREREQS);
  //my_svcudp_destroy(xprt, 0);

}



void
my_svc_getreqset (svc_dispatch disp_fn, fd_set *readfds )
{
  register u_int32_t mask;
  register u_int32_t *maskp;
  register int setsize;
  register int sock;
  register int bit;

  maskp = (u_int32_t *) __FDS_BITS(readfds);
  setsize = _rpc_dtablesize ();
  for (sock = 0; sock < setsize; sock += 32)
    for (mask = *maskp++; (bit = ffs (mask)); mask ^= (1 << (bit - 1)))
      my_svc_getreq_common (disp_fn, sock + bit - 1 );
}


int my_svc_run(svc_dispatch disp_fn)
{
  struct timeval timeout;
  fd_set readfds;

  readfds = svc_fdset;
  timeout.tv_sec = timeout.tv_usec = 2;

  while(1) {

      int i;
      timeout.tv_sec = timeout.tv_usec = 2;
      readfds = svc_fdset;
    i = select(20, &readfds, (fd_set *) 0, (fd_set *) 0, &timeout);
      switch (i) {
        case -1:
          if (errno == EINTR)
        continue;
          perror("svc_run: - select failed");
          break;
        case 0:
          break;
        default:
        /*do the actual service */
          //printf("\nCalling my_svc_getreqset"); fflush(stdout);
          my_svc_getreqset(disp_fn, &readfds );

    }
  }
}
