/***********************************************************************
 *
 * NeST: Network Storage Technologies
 * Copyright (C) 2001-2002, Computer Sciences Department,
 * University of Wisconsin-Madison, WI, and John Bent,
 * Venkateshwaran Venkataramani, and other members of the Condor Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * For contact information, please visit:
 *      http://www.cs.wisc.edu/condor/nest/
 *
 * This particular file was written by Muthian Sivathanu
 ***********************************************************************/


#include <rpc/svc.h>
#include <rpc/auth_unix.h>

#define RQCRED_SIZE 400

struct xp_ops {
    bool_t    (*xp_recv) __PMT ((SVCXPRT *__xprt, struct rpc_msg *__msg));
                /* receive incoming requests */
    enum xprt_stat (*xp_stat) __PMT ((SVCXPRT *__xprt));
                /* get transport status */
    bool_t    (*xp_getargs) __PMT ((SVCXPRT *__xprt, xdrproc_t __xdr_args,
                      caddr_t args_ptr)); /* get arguments */
    bool_t    (*xp_reply) __PMT ((SVCXPRT *__xprt, struct rpc_msg *__msg));
                /* send reply */
    bool_t    (*xp_freeargs) __PMT ((SVCXPRT *__xprt, xdrproc_t __xdr_args,
                       caddr_t args_ptr));
                /* free mem allocated for args */
    void    (*xp_destroy) __PMT ((SVCXPRT *__xprt));
                /* destroy this struct */
};


typedef void (*svc_dispatch)(struct svc_req *, SVCXPRT *);

SVCXPRT *
my_svcudp_create (int sock);

extern "C" {
    void my_svcudp_destroy (SVCXPRT *xprt, int to_close);
}

void
my_svc_getreqset (svc_dispatch disp_fn, fd_set *readfds );

void
my_svc_getreq_common (svc_dispatch disp_fn, const int fd);

int
my_svc_run (svc_dispatch disp_fn);
